import { User, UserExtended, KudosuHistory } from "./user.js";
import { Beatmap, BeatmapExtended, BeatmapDifficultyAttributes, BeatmapPack, Beatmapset, BeatmapsetExtended } from "./beatmap.js";
import { Leader, Match, MatchInfo, MultiplayerScore, PlaylistItem, Room } from "./multiplayer.js";
import { Rulesets, Mod } from "./misc.js";
import { BeatmapUserScore, Score } from "./score.js";
import { Rankings, Spotlight } from "./ranking.js";
import { ChangelogBuild, UpdateStream } from "./changelog.js";
export { User, UserExtended, KudosuHistory };
export { Beatmap, BeatmapExtended, Beatmapset, BeatmapsetExtended };
export { BeatmapUserScore, Score };
export { Room, Leader, PlaylistItem, MultiplayerScore };
export { Rulesets };
export { ChangelogBuild, UpdateStream };
/**
 * Scopes determine what the API instance can do as a user!
 * @remarks "identify" is always implicity provided
 */
type Scope = "chat.read" | "chat.write" | "chat.write_manage" | "delegate" | "forum.write" | "friends.read" | "identify" | "public";
/**
 * Generates a link for users to click on in order to use your application!
 * @param client_id The Client ID, find it at https://osu.ppy.sh/home/account/edit#new-oauth-application
 * @param redirect_uri The specified Application Callback URL, aka where the user will be redirected upon clicking the button to authorize
 * @param scopes What you want to do with/as the user
 * @returns The link people should click on
 */
export declare function generateAuthorizationURL(client_id: number, redirect_uri: string, scopes: Scope[]): string;
/**
 * If the `API` throws an error, it should always be an `APIError`!
 */
export declare class APIError {
    message: string;
    server: string;
    endpoint: string;
    parameters: string;
    /**
     * @param message The reason why things didn't go as expected
     * @param server The server to which the request was sent
     * @param endpoint The type of resource that was requested from the server
     * @param parameters The filters that were used to specify what resource was wanted
     */
    constructor(message: string, server: string, endpoint: string, parameters: string);
}
/**
 * You can create an API instance using its `createAsync` function!
 */
export declare class API {
    client: {
        id: number;
        secret: string;
    };
    /**
     * Should always be "Bearer"
     */
    token_type: string;
    expires: Date;
    access_token: string;
    /**
     * Valid for an unknown amount of time, allows you to get a new token without going through the Authorization Code Grant!
     * Use the API's `refreshToken` function to do that
     */
    refresh_token?: string;
    /**
     * The osu! user id of the user who went through the Authorization Code Grant
     */
    user?: number;
    scopes: Scope[];
    /**
     * (default `none`) Which events should be logged
     */
    verbose: "none" | "errors" | "all";
    /**
     * (default `https://osu.ppy.sh`) The base url of the server where the requests should land
     * @remarks For tokens, requests will be sent to the `oauth/token` route, other requests will be sent to the `api/v2` route
     */
    server: string;
    /**
     * Use the API's `createAsync` instead of the default constructor if you don't have at least an access_token!
     * `createAsync` should always be your way of creating API instances!!
     */
    constructor(client?: {
        id: number;
        secret: string;
    }, token_type?: string, expires?: Date, access_token?: string, scopes?: Scope[], refresh_token?: string, user?: number, verbose?: "none" | "errors" | "all", server?: string);
    /**
     * Use this instead of `console.log` to log any information
     * @param is_error Is the logging happening because of an error?
     * @param to_log Whatever you would put between the parentheses of `console.log()`
     */
    private log;
    private obtainToken;
    /**
     * The normal way to create an API instance! Make sure to `await` it
     * @param client The ID and the secret of your client, can be found on https://osu.ppy.sh/home/account/edit#new-oauth-application
     * @param user If the instance is supposed to represent a user, use their Authorization Code and the Application Callback URL of your application!
     * @returns A promise with an API instance
     */
    static createAsync(client: {
        id: number;
        secret: string;
    }, user?: {
        code: string;
        redirect_uri: string;
    }, verbose?: "none" | "errors" | "all", server?: string): Promise<API>;
    refreshToken(): Promise<boolean>;
    /**
     * @param endpoint What comes in the URL after `api/`
     * @param parameters The things to specify in the request, such as the beatmap_id when looking for a beatmap
     * @param number_try How many attempts there's been to get the data
     * @returns A Promise with either the API's response or `false` upon failing
     */
    private request;
    /**
     * Get extensive user data about the authorized user
     * @param ruleset Defaults to the user's default/favourite Ruleset
     * @scope identify
     */
    getResourceOwner(ruleset?: Rulesets): Promise<UserExtended>;
    /**
     * Get extensive user data about whoever you want!
     * @param user An object with either the id or the username of the user you're trying to get
     * @param ruleset Defaults to the user's default/favourite Ruleset
     */
    getUser(user: {
        id?: number;
        username?: string;
    } | User, ruleset?: Rulesets): Promise<UserExtended>;
    /**
     * Get user data for up to 50 users at once!
     * @param ids An array composed of the ids of the users you want
     */
    getUsers(ids: number[]): Promise<User[]>;
    /**
     * Get "notable" scores from a user
     * @param user The user who set the scores
     * @param type Do you want scores: in the user's top 100, that are top 1 on a beatmap, that have been recently set?
     * @param limit The maximum amount of scores to be returned
     * @param ruleset The Ruleset the scores were made in, defaults to the user's default/favourite Ruleset
     * @param include_fails Do you want scores where the user didn't survive or quit the map? Defaults to false
     * @param offset How many elements that would be at the top of the returned array get skipped (while still filling the array up to the limit)
     */
    getUserScores(user: {
        id: number;
    } | User, type: "best" | "firsts" | "recent", limit?: number, ruleset?: Rulesets, include_fails?: boolean, offset?: number): Promise<Score[]>;
    /**
     * Get data about the activity of a user kudosu-wise!
     * @param user The user in question
     * @param limit The maximum amount of activities in the returned array, defaults to 5
     * @param offset How many elements that would be at the top of the returned array get skipped (while still filling the array up to the limit)
     */
    getUserKudosu(user: {
        id: number;
    } | User, limit?: number, offset?: number): Promise<KudosuHistory[]>;
    /**
     * Get user data of each friend of the authorized user
     * @scope friends.read
     */
    getFriends(): Promise<User[]>;
    /**
     * Get extensive beatmap data about whichever beatmap you want!
     * @param beatmap An object with the id of the beatmap you're trying to get
     */
    getBeatmap(beatmap: {
        id: number;
    } | Beatmap): Promise<BeatmapExtended>;
    /**
     * Get extensive beatmap data for up to 50 beatmaps at once!
     * @param ids An array composed of the ids of the beatmaps you want
     */
    getBeatmaps(ids?: number[]): Promise<BeatmapExtended[]>;
    /**
     * Get various data about the difficulty of a beatmap!
     * @remarks Will ignore the customization of your mods
     * @param beatmap The Beatmap in question
     * @param mods Defaults to No Mod, can be a bitset of mods, an array of mod acronyms ("DT" for DoubleTime), or an array of Mods
     * @param ruleset Defaults to the Ruleset the Beatmap was made for, useful to specify if you want a convert
     * @returns The (beatmap) difficulty attributes of the Beatmap
     * @example ```ts
     * await api.getBeatmapDifficultyAttributes({id: 811925}, ["HR", "HD"])
     * ```
     */
    getBeatmapDifficultyAttributes(beatmap: {
        id: number;
    } | Beatmap, mods?: Mod[] | string[] | number, ruleset?: Rulesets): Promise<BeatmapDifficultyAttributes>;
    /**
     * Get the score on a beatmap made by a specific user (with specific mods and on a specific ruleset if needed)
     * @param beatmap The Beatmap the score was made on
     * @param user The User who made the score
     * @param mods The Mods used to make the score, defaults to any, you can use `["NM"]` to filter out scores with mods
     * @param ruleset The Ruleset used to make the score, useful if it was made on a convert
     * @returns An Object with the position of the score according to the specified Mods and Ruleset, and with the score itself
     */
    getBeatmapUserScore(beatmap: {
        id: number;
    } | Beatmap, user: {
        id: number;
    } | User, mods?: string[], ruleset?: Rulesets): Promise<BeatmapUserScore>;
    /**
     * Get the score on a beatmap made by a specific user (with the possibility to specify if the scores are on a convert)
     * @param beatmap The Beatmap the scores were made on
     * @param user The User who made the scores
     * @param ruleset The Ruleset used to make the scores, defaults to the Ruleset the Beatmap was made for
     */
    getBeatmapUserScores(beatmap: {
        id: number;
    } | Beatmap, user: {
        id: number;
    } | User, ruleset?: Rulesets): Promise<Score[]>;
    /**
     * Get extensive beatmapset data about whichever beatmapset you want!
     * @param beatmap An object with the id of the beatmapset you're trying to get
     */
    getBeatmapset(beatmapset: {
        id: number;
    } | Beatmapset): Promise<BeatmapsetExtended>;
    /**
     * Get data about a BeatmapPack using its tag!
     * @param pack An object with the tag of the beatmappack you're trying to get
     * @remarks Currently in https://osu.ppy.sh/beatmaps/packs, when hovering a pack, its link with its tag should show up on your browser's bottom left
     */
    getBeatmapPack(pack: {
        tag: string;
    } | BeatmapPack): Promise<BeatmapPack>;
    /**
     * Get an Array of up to 100 BeatmapPacks of a specific type!
     * @param type The type of the BeatmapPacks, defaults to "standard"
     */
    getBeatmapPacks(type?: "standard" | "featured" | "tournament" | "loved" | "chart" | "theme" | "artist"): Promise<BeatmapPack[]>;
    getChangelogBuild(stream: string, build: string): Promise<ChangelogBuild>;
    getChangelogListing(format: "html" | "markdown", options?: {
        build_version_range?: {
            from?: string;
            to?: string;
        };
        max_id: number;
        stream: string;
    }): Promise<ChangelogBuild[]>;
    getChangelogStreams(): Promise<UpdateStream[]>;
    getRoom(room: {
        id: number;
    } | Room): Promise<Room>;
    /**
     * Get room data for each room fitting the given criterias
     * @scope public
     */
    getRooms(mode?: "active" | "all" | "ended" | "participated" | "owned"): Promise<Room[]>;
    /**
     * Get the room stats of a user from the room, for each user of that room
     * @scope public
     */
    getRoomLeaderboard(room: {
        id: number;
    } | Room): Promise<Leader[]>;
    /**
     * Get the scores on a specific item of a room
     * @scope public
     * @remarks (2023-11-05) Is currently broken on osu!'s side, gotta love the API not being stable!
     */
    getPlaylistItemScores(item: {
        id: number;
        room_id: number;
    } | PlaylistItem): Promise<MultiplayerScore[]>;
    /**
     * @remarks For multiplayer lobbies from the stable (non-lazer) client, with URLs having `community/matches` or `mp`
     * @param id Can be found at the end of the URL of said match
     */
    getMatch(id: number): Promise<Match>;
    /**
     * Gets the info of the 50 most recently created matches, descending order (most recent is at index 0)
     * @remarks For multiplayer lobbies from the stable (non-lazer) client, with URLs having `community/matches` or `mp`
     */
    getMatches(): Promise<MatchInfo[]>;
    getRanking(ruleset: Rulesets, type: "charts" | "country" | "performance" | "score", filter: "all" | "friends", options?: {
        country?: number;
        spotlight?: {
            id: number;
        } | Spotlight;
        variant?: "4k" | "7k";
    }): Promise<Rankings>;
    getSpotlights(): Promise<Spotlight[]>;
}
