var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import fetch from "node-fetch";
import querystring from "querystring";
import { Rulesets } from "./misc.js";
export { Rulesets };
/**
 * Generates a link for users to click on in order to use your application!
 * @param client_id The Client ID, find it at https://osu.ppy.sh/home/account/edit#new-oauth-application
 * @param redirect_uri The specified Application Callback URL, aka where the user will be redirected upon clicking the button to authorize
 * @param scopes What you want to do with/as the user
 * @returns The link people should click on
 */
export function generateAuthorizationURL(client_id, redirect_uri, scopes) {
    const s = String(scopes).replace(/,/g, "%20");
    return `https://osu.ppy.sh/oauth/authorize?client_id=${client_id}&redirect_uri=${redirect_uri}&scope=${s}&response_type=code`;
}
/**
 * If the `API` throws an error, it should always be an `APIError`!
 */
export class APIError {
    /**
     * @param message The reason why things didn't go as expected
     * @param server The server to which the request was sent
     * @param endpoint The type of resource that was requested from the server
     * @param parameters The filters that were used to specify what resource was wanted
     */
    constructor(message, server, endpoint, parameters) {
        this.message = message;
        this.server = server;
        this.endpoint = endpoint;
        this.parameters = parameters;
    }
}
/**
 * You can create an API instance using its `createAsync` function!
 */
export class API {
    /**
     * Use the API's `createAsync` instead of the default constructor if you don't have at least an access_token!
     * `createAsync` should always be your way of creating API instances!!
     */
    constructor(client, token_type, expires, access_token, scopes, refresh_token, user, verbose = "none", server = "https://osu.ppy.sh") {
        this.client = client !== null && client !== void 0 ? client : { id: 0, secret: "" };
        this.token_type = token_type !== null && token_type !== void 0 ? token_type : "";
        this.expires = expires !== null && expires !== void 0 ? expires : new Date();
        this.access_token = access_token !== null && access_token !== void 0 ? access_token : "";
        this.scopes = scopes !== null && scopes !== void 0 ? scopes : [];
        this.refresh_token = refresh_token;
        this.user = user;
        this.verbose = verbose;
        this.server = server;
    }
    /**
     * Use this instead of `console.log` to log any information
     * @param is_error Is the logging happening because of an error?
     * @param to_log Whatever you would put between the parentheses of `console.log()`
     */
    log(is_error, ...to_log) {
        if (this.verbose === "all" || (this.verbose === "errors" && is_error === true)) {
            console.log("osu!api v2 ->", ...to_log);
        }
    }
    obtainToken(body, api) {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield fetch(`${this.server}/oauth/token`, {
                method: "post",
                headers: {
                    "Accept": "application/json",
                    "Content-Type": "application/json"
                },
                body: JSON.stringify(body)
            });
            const json = yield response.json();
            if (!json.access_token) {
                this.log(true, "Unable to obtain a token! Here's what was received from the API:", json);
                throw new APIError("No token obtained", this.server, "oauth/token", body);
            }
            const token = json.access_token;
            const token_payload = JSON.parse(Buffer.from(token.substring(token.indexOf(".") + 1, token.lastIndexOf(".")), "base64").toString('ascii'));
            if (token_payload.sub && token_payload.sub.length) {
                api.user = Number(token_payload.sub);
            }
            api.scopes = token_payload.scopes;
            api.access_token = token;
            api.token_type = json.token_type;
            if (json.refresh_token) {
                api.refresh_token = json.refresh_token;
            }
            const expiration_date = new Date();
            expiration_date.setSeconds(expiration_date.getSeconds() + json.expires_in);
            api.expires = expiration_date;
            return api;
        });
    }
    /**
     * The normal way to create an API instance! Make sure to `await` it
     * @param client The ID and the secret of your client, can be found on https://osu.ppy.sh/home/account/edit#new-oauth-application
     * @param user If the instance is supposed to represent a user, use their Authorization Code and the Application Callback URL of your application!
     * @returns A promise with an API instance
     */
    static createAsync(client, user, verbose = "none", server = "https://osu.ppy.sh") {
        return __awaiter(this, void 0, void 0, function* () {
            const new_api = new API();
            new_api.client = client;
            new_api.verbose = verbose;
            new_api.server = server;
            const body = {
                client_id: client.id,
                client_secret: client.secret,
                grant_type: user ? "authorization_code" : "client_credentials",
                redirect_uri: user ? user.redirect_uri : null,
                code: user ? user.code : null,
                scope: user ? null : "public"
            };
            const api = yield new_api.obtainToken(body, new_api);
            return api;
        });
    }
    refreshToken() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.refresh_token) {
                return false;
            }
            const body = {
                client_id: this.client.id,
                client_secret: this.client.secret,
                grant_type: "refresh_token",
                refresh_token: this.refresh_token
            };
            const response = yield this.obtainToken(body, this);
            return response ? true : false;
        });
    }
    /**
     * @param endpoint What comes in the URL after `api/`
     * @param parameters The things to specify in the request, such as the beatmap_id when looking for a beatmap
     * @param number_try How many attempts there's been to get the data
     * @returns A Promise with either the API's response or `false` upon failing
     */
    request(method, endpoint, parameters, number_try = 1) {
        return __awaiter(this, void 0, void 0, function* () {
            const max_tries = 5;
            let err = "none";
            let to_retry = false;
            if (parameters !== undefined && method === "get") {
                // If a parameter is an empty string or is undefined, remove it
                for (let i = 0; i < Object.entries(parameters).length; i++) {
                    if (!String(Object.values(parameters)[i]).length || Object.values(parameters)[i] === undefined) {
                        i--;
                        delete parameters[Object.keys(parameters)[i + 1]];
                    }
                }
                // If a parameter is an Array, add "[]" to its name, so the server understands the request properly
                for (let i = 0; i < Object.entries(parameters).length; i++) {
                    if (Array.isArray(Object.values(parameters)[i]) && !Object.keys(parameters)[i].includes("[]")) {
                        i--;
                        parameters[`${Object.keys(parameters)[i + 1]}[]`] = Object.values(parameters)[i + 1];
                        delete parameters[Object.keys(parameters)[i + 1]];
                    }
                }
            }
            const response = yield fetch(`${this.server}/api/v2/${endpoint}?` + (method === "get" && parameters ? querystring.stringify(parameters) : ""), {
                method,
                headers: {
                    "Accept": "application/json",
                    "Accept-Encoding": "gzip",
                    "Content-Type": "application/json",
                    "User-Agent": "osu-api-v2-js (https://github.com/TTTaevas/osu-api-v2-js)",
                    "Authorization": `${this.token_type} ${this.access_token}`
                },
                body: method === "post" ? JSON.stringify(parameters) : null,
            })
                .catch((error) => {
                this.log(true, error.message);
                err = `${error.name} (${error.errno})`;
            });
            if (!response || !response.ok) {
                if (response) {
                    err = response.statusText;
                    if (response.status === 401) {
                        this.log(true, "Server responded with status code 401, maybe you need to do this action as an user?");
                    }
                    else if (response.status === 403) {
                        this.log(true, "Server responded with status code 403, you may lack the necessary scope for this action!");
                    }
                    else if (response.status === 429) {
                        this.log(true, "Server responded with status code 429, you're sending too many requests at once and are getting rate-limited!");
                        to_retry = true;
                    }
                    else {
                        this.log(true, "Server responded with status:", response.status);
                    }
                }
                /**
                 * Under specific circumstances, we want to retry our request automatically
                 * However, spamming the server during the same second in any of these circumstances would be pointless
                 * So we wait for 1 to 5 seconds to make our request, 5 times maximum
                */
                if (to_retry === true && number_try < max_tries) {
                    this.log(true, "Will request again in a few instants...", `(Try #${number_try})`);
                    const to_wait = (Math.floor(Math.random() * (500 - 100 + 1)) + 100) * 10;
                    yield new Promise(res => setTimeout(res, to_wait));
                    return yield this.request(method, endpoint, parameters, number_try + 1);
                }
                throw new APIError(err, `${this.server}/api/v2`, endpoint, JSON.stringify(parameters));
            }
            this.log(false, response.statusText, response.status, { endpoint, parameters });
            return response.json();
        });
    }
    // USER STUFF
    /**
     * Get extensive user data about the authorized user
     * @param ruleset Defaults to the user's default/favourite Ruleset
     * @scope identify
     */
    getResourceOwner(ruleset) {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield this.request("get", "me", { mode: ruleset });
            return correctType(response);
        });
    }
    /**
     * Get extensive user data about whoever you want!
     * @param user An object with either the id or the username of the user you're trying to get
     * @param ruleset Defaults to the user's default/favourite Ruleset
     */
    getUser(user, ruleset) {
        return __awaiter(this, void 0, void 0, function* () {
            const key = user.id !== undefined ? "id" : "username";
            const lookup = user.id !== undefined ? user.id : user.username;
            const mode = ruleset ? `/${Rulesets[ruleset]}` : "";
            const response = yield this.request("get", `users/${lookup}${mode}`, { key });
            return correctType(response);
        });
    }
    /**
     * Get user data for up to 50 users at once!
     * @param ids An array composed of the ids of the users you want
     */
    getUsers(ids) {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield this.request("get", "users", { ids });
            return response.users.map((u) => correctType(u));
        });
    }
    /**
     * Get "notable" scores from a user
     * @param user The user who set the scores
     * @param type Do you want scores: in the user's top 100, that are top 1 on a beatmap, that have been recently set?
     * @param limit The maximum amount of scores to be returned
     * @param ruleset The Ruleset the scores were made in, defaults to the user's default/favourite Ruleset
     * @param include_fails Do you want scores where the user didn't survive or quit the map? Defaults to false
     * @param offset How many elements that would be at the top of the returned array get skipped (while still filling the array up to the limit)
     */
    getUserScores(user, type, limit, ruleset, include_fails = false, offset) {
        return __awaiter(this, void 0, void 0, function* () {
            const mode = ruleset ? Rulesets[ruleset] : undefined;
            const response = yield this.request("get", `users/${user.id}/scores/${type}`, { limit, mode, offset, include_fails: String(Number(include_fails)) });
            return response.map((s) => correctType(s));
        });
    }
    /**
     * Get data about the activity of a user kudosu-wise!
     * @param user The user in question
     * @param limit The maximum amount of activities in the returned array, defaults to 5
     * @param offset How many elements that would be at the top of the returned array get skipped (while still filling the array up to the limit)
     */
    getUserKudosu(user, limit, offset) {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield this.request("get", `users/${user.id}/kudosu`, { limit, offset });
            return response.map((k) => correctType(k));
        });
    }
    /**
     * Get user data of each friend of the authorized user
     * @scope friends.read
     */
    getFriends() {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield this.request("get", "friends");
            return response.map((f) => correctType(f));
        });
    }
    // BEATMAP STUFF
    /**
     * Get extensive beatmap data about whichever beatmap you want!
     * @param beatmap An object with the id of the beatmap you're trying to get
     */
    getBeatmap(beatmap) {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield this.request("get", `beatmaps/${beatmap.id}`);
            return correctType(response);
        });
    }
    /**
     * Get extensive beatmap data for up to 50 beatmaps at once!
     * @param ids An array composed of the ids of the beatmaps you want
     */
    getBeatmaps(ids) {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield this.request("get", "beatmaps", { ids });
            return response.beatmaps.map((b) => correctType(b));
        });
    }
    /**
     * Get various data about the difficulty of a beatmap!
     * @remarks Will ignore the customization of your mods
     * @param beatmap The Beatmap in question
     * @param mods Defaults to No Mod, can be a bitset of mods, an array of mod acronyms ("DT" for DoubleTime), or an array of Mods
     * @param ruleset Defaults to the Ruleset the Beatmap was made for, useful to specify if you want a convert
     * @returns The (beatmap) difficulty attributes of the Beatmap
     * @example ```ts
     * await api.getBeatmapDifficultyAttributes({id: 811925}, ["HR", "HD"])
     * ```
     */
    getBeatmapDifficultyAttributes(beatmap, mods, ruleset) {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield this.request("post", `beatmaps/${beatmap.id}/attributes`, { ruleset_id: ruleset, mods });
            return correctType(response.attributes);
        });
    }
    /**
     * Get the score on a beatmap made by a specific user (with specific mods and on a specific ruleset if needed)
     * @param beatmap The Beatmap the score was made on
     * @param user The User who made the score
     * @param mods The Mods used to make the score, defaults to any, you can use `["NM"]` to filter out scores with mods
     * @param ruleset The Ruleset used to make the score, useful if it was made on a convert
     * @returns An Object with the position of the score according to the specified Mods and Ruleset, and with the score itself
     */
    getBeatmapUserScore(beatmap, user, mods, ruleset) {
        return __awaiter(this, void 0, void 0, function* () {
            const mode = ruleset ? Rulesets[ruleset] : undefined;
            const response = yield this.request("get", `beatmaps/${beatmap.id}/scores/users/${user.id}`, { mods, mode });
            return correctType(response);
        });
    }
    /**
     * Get the score on a beatmap made by a specific user (with the possibility to specify if the scores are on a convert)
     * @param beatmap The Beatmap the scores were made on
     * @param user The User who made the scores
     * @param ruleset The Ruleset used to make the scores, defaults to the Ruleset the Beatmap was made for
     */
    getBeatmapUserScores(beatmap, user, ruleset) {
        return __awaiter(this, void 0, void 0, function* () {
            const mode = ruleset ? Rulesets[ruleset] : undefined;
            const response = yield this.request("get", `beatmaps/${beatmap.id}/scores/users/${user.id}/all`, { mode });
            return response.scores.map((s) => correctType(s));
        });
    }
    /**
     * Get extensive beatmapset data about whichever beatmapset you want!
     * @param beatmap An object with the id of the beatmapset you're trying to get
     */
    getBeatmapset(beatmapset) {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield this.request("get", `beatmapsets/${beatmapset.id}`);
            return correctType(response);
        });
    }
    /**
     * Get data about a BeatmapPack using its tag!
     * @param pack An object with the tag of the beatmappack you're trying to get
     * @remarks Currently in https://osu.ppy.sh/beatmaps/packs, when hovering a pack, its link with its tag should show up on your browser's bottom left
     */
    getBeatmapPack(pack) {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield this.request("get", `beatmaps/packs/${pack.tag}`);
            return correctType(response);
        });
    }
    /**
     * Get an Array of up to 100 BeatmapPacks of a specific type!
     * @param type The type of the BeatmapPacks, defaults to "standard"
     */
    getBeatmapPacks(type = "standard") {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield this.request("get", "beatmaps/packs", { type });
            return correctType(response.beatmap_packs);
        });
    }
    // CHANGELOG STUFF
    getChangelogBuild(stream, build) {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield this.request("get", `changelog/${stream}/${build}`);
            return correctType(response);
        });
    }
    getChangelogListing(format, options) {
        return __awaiter(this, void 0, void 0, function* () {
            const message_format = [format];
            const from = options ? options.build_version_range ? options.build_version_range.from ? options.build_version_range.from : "" : "" : "";
            const to = options ? options.build_version_range ? options.build_version_range.from ? options.build_version_range.from : "" : "" : "";
            const stream = options ? options.stream ? options.stream : "" : "";
            const max_id = options ? options.max_id ? options.max_id : "" : "";
            const response = yield this.request("get", `changelog`, { message_format, from, to, stream, max_id });
            return response.builds.map((b) => correctType(b));
        });
    }
    getChangelogStreams() {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield this.request("get", `changelog`, { max_id: 1 });
            return response.streams.map((s) => correctType(s));
        });
    }
    // MULTIPLAYER STUFF
    getRoom(room) {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield this.request("get", `rooms/${room.id}`);
            return correctType(response);
        });
    }
    /**
     * Get room data for each room fitting the given criterias
     * @scope public
     */
    getRooms(mode = "active") {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield this.request("get", "rooms", { mode });
            return response.map((r) => correctType(r));
        });
    }
    /**
     * Get the room stats of a user from the room, for each user of that room
     * @scope public
     */
    getRoomLeaderboard(room) {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield this.request("get", `rooms/${room.id}/leaderboard`);
            return response.leaderboard.map((l) => correctType(l));
        });
    }
    /**
     * Get the scores on a specific item of a room
     * @scope public
     * @remarks (2023-11-05) Is currently broken on osu!'s side, gotta love the API not being stable!
     */
    getPlaylistItemScores(item) {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield this.request("get", `rooms/${item.room_id}/playlist/${item.id}/scores`);
            return response.scores.map((s) => correctType(s));
        });
    }
    /**
     * @remarks For multiplayer lobbies from the stable (non-lazer) client, with URLs having `community/matches` or `mp`
     * @param id Can be found at the end of the URL of said match
     */
    getMatch(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield this.request("get", `matches/${id}`);
            return correctType(response);
        });
    }
    /**
     * Gets the info of the 50 most recently created matches, descending order (most recent is at index 0)
     * @remarks For multiplayer lobbies from the stable (non-lazer) client, with URLs having `community/matches` or `mp`
     */
    getMatches() {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield this.request("get", "matches");
            return response.matches.map((m) => correctType(m));
        });
    }
    // RANKING STUFF
    getRanking(ruleset, type, filter, options) {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield this.request("get", `rankings/${Rulesets[ruleset]}/${type}`, {
                filter,
                country: options === null || options === void 0 ? void 0 : options.country,
                spotlight: options === null || options === void 0 ? void 0 : options.spotlight,
                variant: options === null || options === void 0 ? void 0 : options.variant
            });
            return correctType(response);
        });
    }
    getSpotlights() {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield this.request("get", "spotlights");
            return response.spotlights.map((s) => correctType(s));
        });
    }
}
/**
 * Some stuff doesn't have the right type to begin with, such as dates, which are being returned as strings, this fixes that
 * @param x Anything, but should be a string, an array that contains a string, or an object which has a string
 * @returns x, but with it (or what it contains) now having the correct type
 */
function correctType(x) {
    if (typeof x === "boolean") {
        return x;
    }
    else if (!isNaN(x)) {
        return x === null ? null : Number(x);
    }
    else if (/^[+-[0-9][0-9]+-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}Z$/.test(x) ||
        /^[+-[0-9][0-9]+-[0-9]{2}-[0-9]{2}$/g.test(x) || /^[+-[0-9][0-9]+-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}\.[0-9]{0,}Z$/g.test(x)) {
        return new Date(x);
    }
    else if (/^[+-[0-9][0-9]+-[0-9]{2}-[0-9]{2} [0-9]{2}:[0-9]{2}:[0-9]{2}$/.test(x)) {
        x += "Z";
        return new Date(x);
    }
    else if (/^[+-[0-9][0-9]+-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}\+[0-9]{2}:[0-9]{2}$/.test(x)) {
        x = x.substring(0, x.indexOf("+")) + "Z";
        return new Date(x);
    }
    else if (Array.isArray(x)) {
        return x.map((e) => correctType(e));
    }
    else if (typeof x === "object" && x !== null) {
        const k = Object.keys(x);
        const v = Object.values(x);
        for (let i = 0; i < k.length; i++) {
            if (k[i] == "name")
                continue; // don't turn names made of numbers into integers
            x[k[i]] = correctType(v[i]);
        }
    }
    return x;
}
